/*----------------------------------------------------------------------
  File    : chtime.c
  Contents: program to time convex hull constructions
  Author  : Christian Borgelt
  History : 2000.10.28 file created
            2000.12.10 several options added
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>
#include "random.h"
#include "chull.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "chtime"
#define DESCRIPTION "time convex hull construction"
#define VERSION     "version 1.1 (14.07.2001)    " \
                    "(c) 2000-2001   Christian Borgelt"

/* --- error codes --- */
#define E_NONE        0         /* no error */
#define E_NOMEM     (-1)        /* not enough memory */
#define E_FOPEN     (-2)        /* file open failed */
#define E_FREAD     (-3)        /* file read failed */
#define E_FWRITE    (-4)        /* file write failed */
#define E_OPTION    (-5)        /* unknown option */
#define E_OPTARG    (-6)        /* missing option argument */
#define E_ARGCNT    (-7)        /* wrong number of arguments */
#define E_REGION    (-8)        /* invalid point region */
#define E_PTSCNT    (-9)        /* invalid number of points */
#define E_PTSINC   (-10)        /* invalid increment */
#define E_STEPS    (-11)        /* invalid number of steps */
#define E_TRIALS   (-12)        /* invalid number of trials */
#define E_UNKNOWN  (-13)        /* unknown error */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {    /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_OPTION   -5 */  "unknown option -%c",
  /* E_OPTARG   -6 */  "missing option argument",
  /* E_ARGCNT   -7 */  "wrong number of arguments",
  /* E_REGION   -8 */  "invalid point region: %c",
  /* E_PTSCNT   -9 */  "invalid number of points: %d",
  /* E_PTSINC  -10 */  "invalid increment: %d",
  /* E_STEPS   -11 */  "invalid number of steps: %d",
  /* E_TRIALS  -12 */  "invalid number of trials: %d",
  /* E_UNKNOWN -13 */  "unknown error",
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static const char *prgname;     /* program name for error messages */
static CHULL *chull = NULL;     /* convex hull structure */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print an error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */

  assert(prgname);              /* check the program name */
  if (code < E_UNKNOWN) code = E_UNKNOWN;
  if (code < 0) {               /* if to report an error, */
    msg = errmsgs[-code];       /* get the error message */
    if (!msg) msg = errmsgs[-E_UNKNOWN];
    fprintf(stderr, "\n%s: ", prgname);
    va_start(args, code);       /* get variable arguments */
    vfprintf(stderr, msg, args);/* print the error message */
    va_end(args);               /* end argument evaluation */
  }
  #ifndef NDEBUG                /* if debugging version */
  if (chull) ch_delete(chull);  /* clean up memory */
  #endif
  #ifdef STORAGE
  showmem("at end of program"); /* check memory usage */
  #endif
  exit(code);                   /* abort the program */
}  /* error() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  int     i, k = 0, n;          /* loop variables, counter */
  char    *s;                   /* to traverse the options */
  char    **optarg = NULL;      /* option argument */
  char    *fname   = NULL;      /* name of the output file */
  int     alg      = CH_SORTED; /* algorithm for the construction */
  int     region   = 'f';       /* point region */
  int     ptscnt   = 0;         /* initial number of points */
  int     ptsinc   = 1000;      /* increment for number of points */
  int     steps    = 10;        /* number of steps  of  ptscnt */
  int     trials   = 10;        /* number of trials per ptscnt */
  clock_t t;                    /* execution time in clock ticks */
  double  x, y, z;              /* coordinates */
  double  avg;                  /* average of execution times */
  FILE    *out;                 /* output file */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] outfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-c       conflict graph algorithm "
                    "(default: sorted points)\n");
    printf("-r#      point region (default: sphere surface)\n");
    printf("         c: cube, s: sphere, f: sphere surface\n");
    printf("-n#      initial number of points "
                    "(default: %d)\n", ptscnt);
    printf("-i#      increment for number of points "
                    "(default: %d)\n", ptsinc);
    printf("-s#      number of steps  (default: %d)\n", steps);
    printf("-t#      number of trials (default: %d)\n", trials);
    printf("outfile  file to write the results to (\"-\": stdout)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse the arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'c': alg    = CH_CGRAPH;              break;
          case 'r': region = (*s) ? *s++ : 'f';      break;
          case 'n': ptscnt = (int)strtol(s, &s, 10); break;
          case 'i': ptsinc = (int)strtol(s, &s, 10); break;
          case 's': steps  = (int)strtol(s, &s, 10); break;
          case 't': trials = (int)strtol(s, &s, 10); break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fname = s;       break;
        default: error(E_ARGCNT); break;
      }                         /* note filename */
    }
  }
  if (optarg)  error(E_OPTARG); /* check option argument */
  if (k != 1)  error(E_ARGCNT); /* check number of arguments */
  if ((region != 'c') && (region != 's') && (region != 'f'))
    error(E_REGION, region);    /* check the point region */
  if (ptscnt <  0) error(E_PTSCNT, ptscnt);  /* check the */
  if (ptsinc <= 0) error(E_PTSINC, ptsinc);  /* various other */
  if (steps  <= 0) error(E_STEPS,  steps);   /* parameters */
  if (trials <= 0) error(E_TRIALS, trials);
  printf("\n");                 /* terminate the startup message */
  switch (region) {             /* evaluate the point region */
    case 'c': region = CH_CUBE;    break;
    case 's': region = CH_SPHERE;  break;
    default : region = CH_SURFACE; break;
  }                             /* get the region identifier */

  /* --- time convex hull construction --- */
  rseed((unsigned)time(NULL));  /* init. random number generator */
  if (!fname || !*fname)        /* if no file name is given, */
    out = stdout;               /* set the standard output, */
  else {                        /* if a  file name is given, */
    out = fopen(fname, "w");    /* open the file */
    if (!out) error(E_FOPEN);   /* for writing the results */
  }
  while (--steps >= 0) {        /* loop over numbers of points */
    avg = 0;                    /* initialize the average proc. time */
    n   = (ptscnt <= 0) ? 4 : ptscnt;
    for (i = trials; --i >= 0;){/* loop over trials with same ptscnt */
      if (chull)                /* if there already is a convex hull, */
        ch_delete(chull);       /* delete it first */
      chull = ch_create();      /* create a new convex hull */
      if (!chull) error(E_NOMEM);
      for (k = n; --k >= 0; ) { /* loop over points */
        ch_rand(region, drand, &x, &y, &z);
        if (ch_point(chull, x, y, z) != 0) error(E_NOMEM);
      }                         /* randomly create ptscnt points */
      t = clock();              /* get the start time */
      if (ch_build(chull, alg) != 0)
        error(E_NOMEM);         /* build the convex hull structure */
      t = clock() -t;           /* and get the processing time */
      avg += (double)t /(double)CLOCKS_PER_SEC;
    }                           /* sum the processing times and */
    avg /= trials;              /* afterwards compute the average */
    fprintf(out, "%6d %6.3f\n", n, avg);
    ptscnt += ptsinc;           /* print the result and */
  }                             /* increment the number of points */
  if ((out != stdout) && (fclose(out) != 0))
    error(E_FWRITE);            /* close the output file */

  #ifndef NDEBUG                /* in the debug version */
  ch_delete(chull);             /* clean up the memory */
  #endif
  #ifdef STORAGE                /* in storage debugging version */
  showmem("at end of program"); /* check memory usage */
  #endif
  return 0;                     /* return 'ok' */
}  /* main() */
