/*----------------------------------------------------------------------
  File    : xchull.c
  Contents: X11 program for convex hull construction
  Author  : Christian Borgelt
  History : 2000.10.19 file created from file xskel.c
            2000.10.22 first version completed
            2000.10.23 Actions > Reset View added
            2000.12.09 visibility of edges added to projection dialog
            2000.12.10 actions menu simplified
            2001.06.15 adapted to redesigned proj3d module
            2001.06.16 double buffering added
            2001.07.13 scaling factor for mouse movement introduced
            2013.08.09 adapted to higher compiler warning level
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"
#include "proj3d.h"
#include "random.h"
#ifndef CH_DRAW
#define CH_DRAW
#endif
#include "chull.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI      3.14159265358979323846  /* \pi */
#endif

/* --- extensions --- */
#define MINXEXT       64        /* minimal x-extension of window */
#define MINYEXT       40        /* minimal y-extension of window */
#define INITXEXT     320        /* initial x-extension of window */
#define INITYEXT     320        /* initial y-extension of window */

/* --- movement modes --- */
#define ROTATE         1        /* change heading and pitch */
#define YAW            2        /* change heading */
#define PITCH          3        /* change pitch angle */
#define ROLL           4        /* change roll  angle */
#define ZOOM           5        /* change distance to the origin */

/* --- drawing operations --- */
#define LINE        0x01        /* draw a line to new point */
#define MARK        0x02        /* mark point with a cross */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_WIDGET     (-6)       /* widget creation failed */
#define E_DIALOG     (-7)       /* cannot create dialog box */
#define E_GRAPH      (-8)       /* cannot create graphics context */
#define E_UNKNOWN    (-9)       /* unkown error */

#define FMT_MAX      256        /* max. len. of error message format */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- error message data --- */
  const char *name, *type;      /* name and type of the error message */
  const char *class;            /* resource class of error message */
  const char *dflt;             /* default message */
  int        argc;              /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Action Function Prototypes
----------------------------------------------------------------------*/
static void a_mode   (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_mouse  (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_load   (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_save   (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_quit   (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_init   (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_next   (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_build  (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_setup  (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_view   (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_redraw (Widget w, XEvent *e, String *s, Cardinal *c);
static void a_about  (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xchull.rsc"           /* fallback resources */

static XtActionsRec actions[] = {
  { "mode",     a_mode   },     /* set movement mode */
  { "mouse",    a_mouse  },     /* mouse movement */
  { "load",     a_load   },     /* File     > Load Points... */
  { "save",     a_save   },     /* File     > Save Points... */
  { "quit",     a_quit   },     /* File     > Quit */
  { "reinit",   a_init   },     /* Actions  > Reinitialize */
  { "next",     a_next   },     /* Actions  > Next Step */
  { "build",    a_build  },     /* Actions  > Build All */
  { "setup",    a_setup  },     /* Actions  > Setup... */
  { "view",     a_view   },     /* Actions  > Reset View */
  { "redraw",   a_redraw },     /* Actions  > Redraw */
  { "about",    a_about  },     /* Help     > About Xchull... */
  { "db_close", db_close } };   /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "Xchull.Error",
                         "no error",                          0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "Xchull.Error",
                         "not enough memory",                 0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "Xchull.Warning",
                         "cannot open file:\n%s",             1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "Xchull.Warning",
                         "read error on file:\n%s",           1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "Xchull.Warning",
                         "write error on file:\n%s",          1 },
  /* E_INIT     -5 */  { "error",   "init",    "Xchull.Error",
                         "initialization failed",             0 },
  /* E_WIDGET   -6 */  { "error",   "widget",  "Xchull.Error",
                         "widget creation failed",            0 },
  /* E_DIALOG   -7 */  { "warning", "dialog",  "Xchull.Warning",
                         "cannot create dialog box",          0 },
  /* E_GRAPH    -8 */  { "error",   "gc",      "Xchull.Error",
                         "cannot create graphics context",    0 },
  /* E_UNKNOWN  -9 */  { "error", "unknown",   "Xchull.Error",
                         "unknown error",                     0 } };

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtAppContext appctx;     /* X11 application context */
static Display *display;        /* display connected to */
static int     screen;          /* screen of the display */
static Window  window;          /* main window */
static Atom    wm_delwin;       /* delete window atom of window mgr. */
static char    fmt[FMT_MAX];    /* format and buffer for messages */
static char    buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget  w_top;           /* top level widget */
static Widget  w_main;          /* main window widget */
static Widget  w_view;          /* viewport widget */
static MENU    *menu;           /* pull down menu */
static GC      gc;              /* graphics context for maze */

/* --- property sheets --- */
static PSHEET  *psh = NULL;     /* parameter setup */

/* --- colors --- */
static unsigned long black;     /* basic colors */
static unsigned long white;

/* --- parameters --- */
struct {                        /* parameter structure */
  int    ptcnt;                 /* number of points */
  int    region;                /* region of points */
  int    alg;                   /* construction algorithm */
  int    visib;                 /* visibility of edges */
} params = { 100, CH_SURFACE, CH_SORTED, CH_SOLID };

static CHULL  *chull  = NULL;   /* convex hull structure */
static int    state   = 0;      /* construction state */
static PROJ3D *proj3d = NULL;   /* 3D to 2D projection */
static int    mode    = 0;      /* movement mode */
static int    mx = 0, my = 0;   /* old mouse coordinates */
static double xoff, yoff;       /* coordinate offsets */
static double xscl, yscl;       /* scaling factors */
static Pixmap buffer  = None;   /* buffer for drawing */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (XtClass(widget) != transientShellWidgetClass)
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close dialog box and */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs -code;          /* get error message data and */
  va_start(va, code);           /* list of variable arguments */
  for (argc = 0; (int)argc < msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  if (strcmp(msg->name, "error") == 0)  /* if name is 'error' */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is 'warning' */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_nop, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_load (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Points...' callback */
  int   res;                    /* result of ch_load() */
  CHULL *tmp;                   /* loaded convex hull structure */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  tmp = ch_create();            /* create a convex hull */
  if (!tmp) { error(E_NOMEM); return; }
  res = ch_load(tmp, fname);    /* load points of convex hull */
  if (res < 0) {                /* and check for success */
    if (res >= -1) error(E_NOMEM);
    else error((res < -2) ? E_FREAD : E_FOPEN, fname);
    ch_delete(tmp); return;     /* on an error delete the created */
  }                             /* convex hull structure and abort */
  if (chull) ch_delete(chull);  /* delete an existing convex hull */
  chull = tmp;                  /* and set the loaded structure */
  state = 0;                    /* and the construction state */
  mn_enable(menu, MN_ITEMS, 3); /* reset the menu state and */
  mn_info(menu, "");            /* and clear the processing time */
  a_view  (NULL,NULL,NULL,NULL);/* reset the viewing parameters */
  a_redraw(NULL,NULL,NULL,NULL);/* redraw the window contents */
}  /* dcb_load() */

/*--------------------------------------------------------------------*/

static void dcb_save (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Save Points...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  if (ch_save(chull, fname) != 0)
    error(E_FWRITE, fname);     /* save points of convex hull */
}  /* dcb_save() */

/*--------------------------------------------------------------------*/

static void dcb_setup (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Setup...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */

  params.ptcnt = atoi(psh_getval(psh, "ptcnt"));
  if (params.ptcnt < 4) params.ptcnt = 4;
  switch ((long)psh_getval(psh, "region")) {
    case 1 : params.region = CH_SPHERE;  break;
    case 2 : params.region = CH_SURFACE; break;
    default: params.region = CH_CUBE;    break;
  }                             /* get the point region */
  switch ((long)psh_getval(psh, "alg")) {
    case 1 : params.alg    = CH_CGRAPH;  break;
    default: params.alg    = CH_SORTED;  break;
  }                             /* get the algorithm */
  switch ((long)psh_getval(psh, "visib")) {
    case 1 : params.visib  = CH_SOLID;   break;
    default: params.visib  = CH_WFRAME;  break;
  }                             /* get the visibility mode */
  a_redraw(NULL,NULL,NULL,NULL);/* redraw the window contents */
}  /* dcb_setup() */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_load   (Widget widget, XtPointer client, XtPointer call)
{ a_load(NULL, NULL, NULL, NULL); }

static void mcb_save   (Widget widget, XtPointer client, XtPointer call)
{ a_save(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer client, XtPointer call)
{ a_quit(NULL, NULL, NULL, NULL); }

static void mcb_reinit (Widget widget, XtPointer client, XtPointer call)
{ a_init(NULL, (XEvent*)1, NULL, NULL); }

static void mcb_next   (Widget widget, XtPointer client, XtPointer call)
{ a_next(NULL, NULL, NULL, NULL); }

static void mcb_build  (Widget widget, XtPointer client, XtPointer call)
{ a_build(NULL, NULL, NULL, NULL); }

static void mcb_setup  (Widget widget, XtPointer client, XtPointer call)
{ a_setup(NULL, NULL, NULL, NULL); }

static void mcb_view   (Widget widget, XtPointer client, XtPointer call)
{ a_view(NULL, (XEvent*)1, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer client, XtPointer call)
{ a_redraw(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer client, XtPointer call)
{ a_about(NULL, NULL, NULL, NULL); }

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void a_mode (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- set movement mode */
  if (!c || (*c != 1)) return;  /* check for an argument */
  if      (strcmp(*s, "rotate") == 0) mode = ROTATE;
  else if (strcmp(*s, "yaw")    == 0) mode = YAW;
  else if (strcmp(*s, "pitch")  == 0) mode = PITCH;
  else if (strcmp(*s, "roll")   == 0) mode = ROLL;
  else if (strcmp(*s, "zoom")   == 0) mode = ZOOM;
  else                                mode = 0;
  mx = e->xbutton.x;            /* note the movement mode and */
  my = e->xbutton.y;            /* the current mouse position */
  p3d_save(proj3d);             /* and save the current view */
}  /* a_mode() */

/*--------------------------------------------------------------------*/

static void a_mouse (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- evaluate mouse movement */
  int    dx, dy;                /* pixels mouse was moved */
  double sf;                    /* scaling factor */

  if (mode <= 0) return;        /* check for a movement mode */
  dx = e->xbutton.x -mx;        /* get the number of pixels */
  dy = e->xbutton.y -my;        /* the mouse was moved */
  sf = 75.0 /xscl;              /* compute the scaling factor */
  switch (mode) {               /* according to the movement mode */
    case ROTATE: if ((dx == 0) && (dy == 0)) return;
                 p3d_restore(proj3d);
                 p3d_pitch(proj3d, dy*sf*(M_PI/180), P3D_BOTH);
                 p3d_yaw  (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case YAW:    if (dx == 0) return;
                 p3d_restore(proj3d);
                 p3d_yaw  (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case PITCH:  if (dy == 0) return;
                 p3d_restore(proj3d);
                 p3d_pitch(proj3d, dy*sf*(M_PI/180), P3D_BOTH); break;
    case ROLL:   if (dx == 0) return;
                 p3d_restore(proj3d);
                 p3d_roll (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case ZOOM:   if (dy == 0) return;
                 p3d_restore(proj3d);
                 p3d_move (proj3d, 0, dy*0.02, 0);              break;
  }                             /* change the viewing parameters */
  a_redraw(NULL,NULL,NULL,NULL);/* and redraw the convex hull */
}  /* a_mouse() */

/*--------------------------------------------------------------------*/

static void a_load (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*load"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_load) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* a_load() */

/*--------------------------------------------------------------------*/

static void a_save (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Save... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*save"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_save) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* a_save() */

/*--------------------------------------------------------------------*/

static void a_quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  #ifndef NDEBUG                /* in debugging version */
  if (chull) ch_delete(chull);  /* delete convex hull structure */
  #endif
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate program */
}  /* a_quit() */

/*--------------------------------------------------------------------*/

static void a_setup (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Setup... */
  int  i;                       /* buffer for menu index */
  char buf[32];                 /* buffer for number */

  if (!psh) {                   /* if property sheet not created yet */
    psh = psh_create("setup", w_top, 1);
    if (!psh) return;           /* create a property sheet */
    if ((psh_addline(psh, "lptcnt")              != 0)
    ||  (psh_additem(psh, "ptcnt",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh, "lregion")             != 0)
    ||  (psh_additem(psh, "region", PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh, "region", "cube")      != 0)
    ||  (psh_menuadd(psh, "region", "sphere")    != 0)
    ||  (psh_menuadd(psh, "region", "surface")   != 0)
    ||  (psh_addline(psh, "lalg")                != 0)
    ||  (psh_additem(psh, "alg",    PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh, "alg",    "sorted")    != 0)
    ||  (psh_menuadd(psh, "alg",    "cgraph")    != 0)
    ||  (psh_addline(psh, "lvisib")              != 0)
    ||  (psh_additem(psh, "visib",  PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh, "visib",  "wframe")    != 0)
    ||  (psh_menuadd(psh, "visib",  "solid")     != 0)) {
      psh_delete(psh); error(E_WIDGET); }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%d", params.ptcnt);
  psh_setval(psh, "ptcnt", buf);/* set the number of points */
  switch (params.region) {      /* map the point region parameter */
    case CH_SPHERE : i = 1; break;
    case CH_SURFACE: i = 2; break;
    default        : i = 0; break;
  }                             /* set the point region */
  psh_setval(psh, "region", (const char*)(long)i);
  switch (params.alg) {         /* map the algorithm parameter */
    case CH_CGRAPH : i = 1; break;
    default        : i = 0; break;
  }                             /* set the algorithm */
  psh_setval(psh, "alg",    (const char*)(long)i);
  switch (params.visib) {       /* map the visibility parameter */
    case CH_SOLID  : i = 1; break;
    default        : i = 0; break;
  }                             /* set the visibility mode */
  psh_setval(psh, "visib",  (const char*)(long)i);
  psh_handle(psh, dcb_setup);   /* handle the dialog */
}  /* a_setup() */

/*--------------------------------------------------------------------*/

static void a_init (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Action > Reinitialize */
  int    i;                     /* loop variable */
  double x, y, z;               /* coordinates of random points */

  if (chull) ch_delete(chull);  /* delete existing structure */
  chull = ch_create();          /* create a new convex hull structure */
  if (!chull) { error(E_NOMEM); return; }
  for (i = params.ptcnt; --i >= 0; ) {   /* generate random points */
    ch_rand(params.region, drand, &x, &y, &z);
    if (ch_point(chull, x, y, z) != 0) {
      error(E_NOMEM); ch_delete(chull); chull = NULL; return; }
  }                             /* add the points to the structure */
  state = 0;                    /* and the construction state */
  a_view(NULL,NULL,NULL,NULL);  /* reset the viewing parameters */
  if (menu) {                   /* update the menu bar */
    mn_enable(menu, MN_ITEMS, 3); mn_info(menu, ""); }
  if (e) a_redraw(NULL,NULL,NULL,NULL);
}  /* a_init() */               /* redraw the window contents */

/*--------------------------------------------------------------------*/

static void a_next (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Next Point */
  int r;                        /* result of ch_next() */

  if (state > 1) return;        /* check the construction state */
  if (state > 0)                /* if construction is already started */
    r = ch_next(chull);         /* process the next point */
  else                          /* if construction is not started, */
    r = ch_start(chull, params.alg);     /* start the construction */
  if (r < 0) error(E_NOMEM);    /* check for an error and */
  state = (r != 0) ? 2 : 1;     /* update the construction state */
  if (state > 1)                /* and the menu */
    mn_enable(menu, MN_ITEMS, 1);
  a_redraw(NULL,NULL,NULL,NULL);/* redraw the window contents */
}  /* a_next() */

/*--------------------------------------------------------------------*/

static void a_build (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Build All */
  char    buf[32];              /* buffer for timer output */
  clock_t t;                    /* starting time */
  int     r;                    /* result of ch_build() */
  double  time;                 /* processing time */

  if (state > 1) return;        /* check the construction state */
  t = clock();                  /* get the starting time */
  r = ch_build(chull, params.alg); /* build the complex hull */
  t = clock() -t;               /* compute the running time */
  state = 2;                    /* update the construction state */
  mn_enable(menu, MN_ITEMS, 1); /* disable some menu entries */
  if (r < 0) { error(E_NOMEM); return; }
  printf("\nconvex hull construction statistics:\n");
  ch_stats(chull);              /* print construction statistics */
  time = (double)t /(double)CLOCKS_PER_SEC;
  printf("construction time       : %.2fs\n", time);
  a_redraw(NULL,NULL,NULL,NULL);/* redraw the window contents */
  sprintf(buf, "%.2fs", time);  /* format and */
  mn_info(menu, buf);           /* display the execution time */
}  /* a_build() */

/*--------------------------------------------------------------------*/

static void a_view (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Reset View */
  p3d_eye  (proj3d, 0, -6, 0);  /* initialize eye, direction of view, */
  p3d_view (proj3d, 0,  0, 0);  /* and the viewing frustrum */
  p3d_frust(proj3d, M_PI/6, 1e-6, 1e6);
  if (e) a_redraw(NULL,NULL,NULL,NULL);
}  /* a_view() */               /* redraw the window contents */

/*--------------------------------------------------------------------*/

static void drawpt (double x, double y, int op)
{                               /* --- move to/draw a line to a point */
  static int px = 0, py = 0;    /* coordinates of preceding point */
  int        cx, cy;            /* coordinates of current point */

  cx = (int)(xoff +x *xscl);    /* transform the coordinates */
  cy = (int)(yoff -y *yscl);    /* w.r.t. the widget's size */
  if (op & 1)                   /* if to draw a line to new point */
    XDrawLine(display, buffer, gc, px, py, cx, cy);
  if (op & 2) {                 /* if to mark new point */
    XDrawLine(display, buffer, gc, cx-1, cy, cx+1, cy);
    XDrawLine(display, buffer, gc, cx, cy-1, cx, cy+1);
  }
  px = cx; py = cy;             /* note coordinates of current point */
}  /* drawpt() */

/*--------------------------------------------------------------------*/

static void a_redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  static Dimension owd  = 0;    /* old window width */
  static Dimension oht  = 0;    /* old window height */
  Dimension  wd, ht;            /* widget width and height */
  XEvent     de;                /* dummy event */
  XRectangle rc;                /* clipping rectangle */

  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  XtVaGetValues(w_view, XtNwidth, &wd, XtNheight, &ht, NULL);
  if (e && (e->type == Expose)  /* if the window was exposed, */
  &&  (wd == owd) && (ht == oht)) {        /* but not resized */
    rc.x      = (short)e->xexpose.x;       /* build and set */
    rc.y      = (short)e->xexpose.y;       /* clipping rectangle */
    rc.width  = (unsigned short)e->xexpose.width;
    rc.height = (unsigned short)e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted); }
  else {                        /* if the window was resized */
    rc.x = rc.y = 0;            /* or this is a forced redraw, */
    rc.width  = wd;             /* redraw all of the window */
    rc.height = ht;             /* (set full window rectangle) */
    while (XCheckTypedWindowEvent(display, window, Expose, &de));
    mn_resize(menu, wd);        /* remove all exposure events */
  }                             /* and resize the menu bar */
  if ((wd != owd) || (ht != oht)/* if the widget extensions changed, */
  ||  (buffer == None)) {       /* a new buffer has to be allocated */
    if ((buffer != None) && (buffer != window))
      XFreePixmap(display, buffer);
    buffer = XCreatePixmap(display, window, wd, ht,
                          (unsigned)XDefaultDepth(display, screen));
    if (buffer == None) buffer = window;
  }                             /* if no buffer can be allocated, */
  owd = wd; oht = ht;           /* note (new) window dimensions */
  xoff = 0.5*wd; yoff = 0.5*ht;
  xscl = 0.5*(wd-1);            /* compute the coordinate offsets */
  yscl = 0.5*(ht-1);            /* and the scaling factors */
  XSetForeground(display, gc, white);
  XFillRectangle(display, buffer, gc, 0, 0, wd, ht);
  XSetForeground(display, gc, black);
  if (chull)                    /* draw the convex hull */
    ch_draw(chull, params.visib, proj3d, drawpt);
  if (buffer != window)         /* copy buffer contents to window */
    XCopyArea(display, buffer, window, gc, 0, 0, wd, ht, 0, 0);
  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);      /* and clear clipping mask */
}  /* a_redraw() */

/*--------------------------------------------------------------------*/

static void a_about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, 0, dcb_nop) != 0)
    error(E_DIALOG);            /* show 'about' dialog box */
}  /* a_about() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg      args[8];             /* widget arguments */
  Cardinal n;                   /* argument counter */

  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);
  rseed((unsigned)time(NULL));  /* init. random number generator */
  proj3d = p3d_create();
  if (!proj3d) error(E_NOMEM);  /* create a projection and init. */
  a_init(NULL,NULL,NULL,NULL);  /* viewing params. and convex hull */

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "Xchull",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET); /* (form for window layout) */
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",     0, 0,          0) != 0)
  ||  (mn_additem (menu, "load",     0, mcb_load,   0) != 0)
  ||  (mn_additem (menu, "save",     0, mcb_save,   0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "quit",     0, mcb_quit,   0) != 0)
  ||  (mn_addtitle(menu, "actions",  0, 0,          0) != 0)
  ||  (mn_additem (menu, "reinit",   0, mcb_reinit, 0) != 0)
  ||  (mn_additem (menu, "next",     2, mcb_next,   0) != 0)
  ||  (mn_additem (menu, "build",    2, mcb_build,  0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "setup",    0, mcb_setup,  0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "view",     0, mcb_view,   0) != 0)
  ||  (mn_additem (menu, "redraw",   0, mcb_redraw, 0) != 0)
  ||  (mn_addtitle(menu, "help",     0, 0,          0) != 0)
  ||  (mn_additem (menu, "about",    0, mcb_about,  0) != 0)
  ||  (mn_addtitle(menu, NULL,       0, 0,          0) != 0))
    error(E_WIDGET);
  mn_resize(menu, INITXEXT);    /* set width of menu bar */
  mn_enable(menu, MN_ITEMS, 3); /* enable all menu items */

  /* --- create viewport --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);      n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);       n++;
  XtSetArg(args[n], XtNright,     XtChainRight);     n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom);    n++;
  XtSetArg(args[n], XtNresizable, True);             n++;
  w_view = XtCreateManagedWidget("view",
             coreWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */
  XtVaSetValues(w_view, XtNwidth,     INITXEXT,
                        XtNheight,    INITYEXT, NULL);
  XtVaSetValues(w_top,  XtNminWidth,  MINXEXT,
                        XtNminHeight, MINYEXT,  NULL);
                                /* set window size */
  XtOverrideTranslations(w_view,/* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));

  /* --- install actions, accelerators and translations --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);

  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_view);   /* and drawable for window */

  if (db_alert(w_top, dcb_nop, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- allocate colors --- */
  screen = DefaultScreen(display);
  white  = WhitePixel(display, screen);
  black  = BlackPixel(display, screen);

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top),
    "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  /* --- evaluate command line arguments --- */
  if (argc > 1) dcb_load(NULL, (XtPointer)1, (XtPointer)argv[1]);

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return 'ok' */
}  /* main() */
